import logging
import os
from typing import (
    Any,
    Optional,
)

from galaxy.tool_util.verify.test_data import TestDataResolver
from galaxy.util.unittest import TestCase
from galaxy_test.base.env import (
    setup_keep_outdir,
    target_url_parts,
)

log = logging.getLogger(__name__)


class FunctionalTestCase(TestCase):
    """Base class for tests targetting actual Galaxy servers.

    Subclass should override galaxy_driver_class if a Galaxy server
    needs to be launched to run the test, this base class assumes a
    server is already running.
    """

    galaxy_driver_class: Optional[type] = None
    host: str
    port: Optional[str]
    url: str
    keepOutdir: str
    test_data_resolver: TestDataResolver
    _test_driver: Optional[Any]

    def setUp(self) -> None:
        self.host, self.port, self.url = target_url_parts()
        server_wrapper = (
            self._test_driver and self._test_driver.server_wrappers and self._test_driver.server_wrappers[0]
        )
        if server_wrapper:
            self.host = server_wrapper.host
            self.port = server_wrapper.port
            self.url = f"http://{self.host}:{self.port}{server_wrapper.prefix.rstrip('/')}/"
        self.test_data_resolver = TestDataResolver()
        self.keepOutdir = setup_keep_outdir()

    @classmethod
    def setUpClass(cls):
        """Configure and start Galaxy for a test."""
        cls._test_driver = None

        if cls.galaxy_driver_class is not None and not os.environ.get("GALAXY_TEST_ENVIRONMENT_CONFIGURED"):
            cls._test_driver = cls.galaxy_driver_class()
            cls._test_driver.setup(config_object=cls)

    @classmethod
    def tearDownClass(cls):
        """Shutdown Galaxy server and cleanup temp directory."""
        if cls._test_driver:
            cls._test_driver.tear_down()

    def get_filename(self, filename: str) -> str:
        # No longer used by tool tests - drop if isn't used else where.
        return self.test_data_resolver.get_filename(filename)



"""Test decorators for test methods meant to be run on external Galaxy instances.

These decorators should not be needed for integration tests, unit tests, etc... but
are appropriate for the API and Selenium tests that are meant to be executable on
external Galaxies.

Running this class of tests should implicitly come with the expectation that new
jobs, workflows, and datasets will be created. But tests creating new published objects,
histories, libraries, etc... should be annotated ideally.
"""
import os
import unittest
from functools import wraps
from typing import Union

import pytest
from typing_extensions import Literal

KnownRequirementT = Union[
    Literal["admin"],
    Literal["celery"],
    Literal["new_history"],
    Literal["new_library"],
    Literal["new_published_objects"],
    Literal["new_user"],
]


def has_requirement(method, tag: KnownRequirementT):
    try:
        # Skipping B009 (use of getattr with constant attribute) in
        # this next line because method may be a bound class method
        # and direct attribute access appends the class name and such.
        return tag in getattr(method, "__required_galaxy_features")  # noqa: B009
    except AttributeError:
        return False


def using_requirement(tag: KnownRequirementT):
    """At runtime, indicate we're using a Galaxy feature.

    This allows runtime test skips if the appropriate environment variable is set.
    """
    requirement = f"requires_{tag}"
    skip_environment_variable = f"GALAXY_TEST_SKIP_IF_{requirement.upper()}"
    env_value = os.environ.get(skip_environment_variable, "0")
    if env_value != "0":
        raise unittest.SkipTest(f"[{env_value}] Skipping due to {skip_environment_variable} being set to {env_value}")


def _attach_requirements(method, tag: KnownRequirementT):
    requirement = f"requires_{tag}"
    if not hasattr(method, "__required_galaxy_features"):
        method.__required_galaxy_features = []
    method.__required_galaxy_features.append(tag)
    getattr(pytest.mark, requirement)(method)


def _wrap_method_with_galaxy_requirement(method, tag: KnownRequirementT):
    _attach_requirements(method, tag)

    @wraps(method)
    def wrapped_method(*args, **kwargs):
        using_requirement(tag)
        return method(*args, **kwargs)

    return wrapped_method


def requires_new_history(method):
    return _wrap_method_with_galaxy_requirement(method, "new_history")


def requires_new_user(method):
    method = _wrap_method_with_galaxy_requirement(method, "new_user")
    method = _wrap_method_with_galaxy_requirement(method, "admin")
    return method


def requires_new_library(method):
    method = _wrap_method_with_galaxy_requirement(method, "new_library")
    method = _wrap_method_with_galaxy_requirement(method, "admin")
    return method


def requires_new_published_objects(method):
    method = _wrap_method_with_galaxy_requirement(method, "new_published_objects")


def requires_admin(method):
    return _wrap_method_with_galaxy_requirement(method, "admin")


def requires_celery(method):
    return _wrap_method_with_galaxy_requirement(method, "celery")


__all__ = (
    "has_requirement",
    "requires_admin",
    "requires_celery",
    "requires_new_history",
    "requires_new_library",
    "requires_new_published_objects",
    "requires_new_user",
    "using_requirement",
)



"""Just constants useful for testing across test types."""

# Following constants used by upload tests.
ONE_TO_SIX_WITH_SPACES = "1 2 3\n4 5 6\n"
ONE_TO_SIX_WITH_TABS = "1\t2\t3\n4\t5\t6\n"
ONE_TO_SIX_ON_WINDOWS = "1\t2\t3\r4\t5\t6\r"
ONE_TO_SIX_WITH_SPACES_ON_WINDOWS = "1 2 3\r4 5 6\r"
ONE_TO_SIX_WITH_TABS_NO_TRAILING_NEWLINE = "1\t2\t3\n4\t5\t6"



import base64
import os
import random
import string
from typing import (
    Dict,
    Optional,
)

DEFAULT_GALAXY_MASTER_API_KEY = "TEST123"
DEFAULT_GALAXY_USER_API_KEY = None

DEFAULT_TEST_USER = "user@bx.psu.edu"
DEFAULT_ADMIN_TEST_USER = "test@bx.psu.edu"
DEFAULT_OTHER_USER = "otheruser@bx.psu.edu"  # A second user for API testing.

TEST_USER = os.environ.get("GALAXY_TEST_USER_EMAIL", DEFAULT_TEST_USER)
ADMIN_TEST_USER = os.environ.get("GALAXY_TEST_ADMIN_USER_EMAIL", DEFAULT_ADMIN_TEST_USER)
OTHER_USER = os.environ.get("GALAXY_TEST_OTHER_USER_EMAIL", DEFAULT_OTHER_USER)


def get_admin_api_key() -> str:
    """Test admin API key to use for functional tests.

    This key should be configured as a admin API key and should be able
    to create additional users and keys.
    """
    for key in ["GALAXY_CONFIG_MASTER_API_KEY", "GALAXY_CONFIG_OVERRIDE_MASTER_API_KEY"]:
        value = os.environ.get(key, None)
        if value:
            return value
    return DEFAULT_GALAXY_MASTER_API_KEY


def get_user_api_key() -> Optional[str]:
    """Test user API key to use for functional tests.

    If set, this should drive API based testing - if not set an admin API key will
    be used to create a new user and API key for tests.
    """
    return os.environ.get("GALAXY_TEST_USER_API_KEY", DEFAULT_GALAXY_USER_API_KEY)


def baseauth_headers(username: str, password: str) -> Dict[str, str]:
    unencoded_credentials = f"{username}:{password}"
    authorization = base64.b64encode(unencoded_credentials.encode("utf-8")).decode("utf-8")
    headers = {
        "Authorization": authorization,
    }
    return headers


def random_name(prefix: Optional[str] = None, suffix: Optional[str] = None, len: int = 10) -> str:
    return "{}{}{}".format(
        prefix or "",
        "".join(random.choice(string.ascii_lowercase + string.digits) for _ in range(len)),
        suffix or "",
    )



def check_example_1(hdca, dataset_populator):
    assert hdca["collection_type"] == "list"
    assert hdca["element_count"] == 2

    first_dce = hdca["elements"][0]
    first_hda = first_dce["object"]
    assert first_hda["hid"] > 3


def check_example_2(hdca, dataset_populator):
    assert hdca["collection_type"] == "list:list"
    assert hdca["element_count"] == 2
    first_collection_level = hdca["elements"][0]
    assert first_collection_level["element_type"] == "dataset_collection"
    second_collection_level = first_collection_level["object"]
    assert second_collection_level["collection_type"] == "list"
    assert second_collection_level["elements"][0]["element_type"] == "hda"


def check_example_3(hdca, dataset_populator):
    assert hdca["collection_type"] == "list"
    assert hdca["element_count"] == 2
    first_element = hdca["elements"][0]
    assert first_element["element_identifier"] == "test0forward"


def check_example_4(hdca, dataset_populator):
    assert hdca["collection_type"] == "list:list"
    assert hdca["element_count"] == 2
    first_collection_level = hdca["elements"][0]
    assert first_collection_level["element_identifier"] == "single", hdca
    assert first_collection_level["element_type"] == "dataset_collection"
    second_collection_level = first_collection_level["object"]
    assert "elements" in second_collection_level, hdca
    assert len(second_collection_level["elements"]) == 1, hdca
    i1_element = second_collection_level["elements"][0]
    assert "object" in i1_element, hdca
    assert "element_identifier" in i1_element
    assert i1_element["element_identifier"] == "i1", hdca
    assert len(i1_element["object"]["tags"]) == 0


def check_example_5(hdca, dataset_populator):
    assert hdca["collection_type"] == "list:list"
    assert hdca["element_count"] == 2
    first_collection_level = hdca["elements"][0]
    assert first_collection_level["element_identifier"] == "single", hdca
    assert first_collection_level["element_type"] == "dataset_collection"
    second_collection_level = first_collection_level["object"]
    assert "elements" in second_collection_level, hdca
    assert len(second_collection_level["elements"]) == 1, hdca
    i1_element = second_collection_level["elements"][0]
    assert "object" in i1_element, hdca
    assert "element_identifier" in i1_element
    assert i1_element["element_identifier"] == "i1", hdca
    tags = i1_element["object"]["tags"]
    assert len(tags) > 0
    assert "group:single" in tags, tags
    assert "i1" in tags, tags


def check_example_6(hdca, dataset_populator):
    assert hdca["collection_type"] == "list"
    assert hdca["element_count"] == 3
    i1_element = hdca["elements"][0]
    assert "object" in i1_element, hdca
    assert "element_identifier" in i1_element
    assert i1_element["element_identifier"] == "i1", hdca
    tags = i1_element["object"]["tags"]
    assert len(tags) == 2
    assert "random" in tags
    assert "group:type:single" in tags


EXAMPLE_1 = {
    "rules": {
        "rules": [
            {
                "type": "add_column_metadata",
                "value": "identifier0",
            }
        ],
        "mapping": [
            {
                "type": "list_identifiers",
                "columns": [0],
            }
        ],
    },
    "test_data": {
        "type": "list",
        "elements": [
            {
                "identifier": "i1",
                "contents": "0",
                "class": "File",
            },
            {
                "identifier": "i2",
                "contents": "1",
                "class": "File",
            },
        ],
    },
    "check": check_example_1,
    "output_hid": 6,
}


EXAMPLE_2 = {
    "rules": {
        "rules": [
            {
                "type": "add_column_metadata",
                "value": "identifier0",
            },
            {
                "type": "add_column_metadata",
                "value": "identifier0",
            },
        ],
        "mapping": [
            {
                "type": "list_identifiers",
                "columns": [0, 1],
            }
        ],
    },
    "test_data": {
        "type": "list",
        "elements": [
            {
                "identifier": "i1",
                "contents": "0",
                "class": "File",
            },
            {
                "identifier": "i2",
                "contents": "1",
                "class": "File",
            },
        ],
    },
    "check": check_example_2,
    "output_hid": 6,
}

# Flatten
EXAMPLE_3 = {
    "rules": {
        "rules": [
            {
                "type": "add_column_metadata",
                "value": "identifier0",
            },
            {
                "type": "add_column_metadata",
                "value": "identifier1",
            },
            {
                "type": "add_column_concatenate",
                "target_column_0": 0,
                "target_column_1": 1,
            },
        ],
        "mapping": [
            {
                "type": "list_identifiers",
                "columns": [2],
            }
        ],
    },
    "test_data": {
        "type": "list:paired",
        "elements": [
            {
                "identifier": "test0",
                "elements": [
                    {"identifier": "forward", "class": "File", "contents": "TestData123"},
                    {"identifier": "reverse", "class": "File", "contents": "TestData123"},
                ],
            }
        ],
    },
    "check": check_example_3,
    "output_hid": 6,
}

# Nesting with group tags.
EXAMPLE_4 = {
    "rules": {
        "rules": [
            {
                "type": "add_column_metadata",
                "value": "identifier0",
            },
            {"type": "add_column_group_tag_value", "value": "type", "default_value": "unused"},
        ],
        "mapping": [
            {
                "type": "list_identifiers",
                "columns": [1, 0],
            }
        ],
    },
    "test_data": {
        "type": "list",
        "elements": [
            {"identifier": "i1", "contents": "0", "class": "File", "tags": ["random", "group:type:single"]},
            {"identifier": "i2", "contents": "1", "class": "File", "tags": ["random", "group:type:paired"]},
            {"identifier": "i3", "contents": "2", "class": "File", "tags": ["random", "group:type:paired"]},
        ],
    },
    "check": check_example_4,
    "output_hid": 8,
}


EXAMPLE_5 = {
    "rules": {
        "rules": [
            {
                "type": "add_column_metadata",
                "value": "identifier0",
            },
            {"type": "add_column_group_tag_value", "value": "type", "default_value": "unused"},
        ],
        "mapping": [
            {
                "type": "list_identifiers",
                "columns": [1, 0],
            },
            {
                "type": "group_tags",
                "columns": [1],
            },
            {
                "type": "tags",
                "columns": [0],
            },
        ],
    },
    "test_data": {
        "type": "list",
        "elements": [
            {"identifier": "i1", "contents": "0", "class": "File", "tags": ["random", "group:type:single"]},
            {"identifier": "i2", "contents": "1", "class": "File", "tags": ["random", "group:type:paired"]},
            {"identifier": "i3", "contents": "2", "class": "File", "tags": ["random", "group:type:paired"]},
        ],
    },
    "check": check_example_5,
    "output_hid": 8,
}


EXAMPLE_6 = {
    "rules": {
        "rules": [
            {
                "type": "add_column_metadata",
                "value": "identifier0",
            },
            {
                "type": "add_column_metadata",
                "value": "tags",
            },
        ],
        "mapping": [
            {
                "type": "list_identifiers",
                "columns": [0],
            },
            {
                "type": "tags",
                "columns": [1],
            },
        ],
    },
    "test_data": {
        "type": "list",
        "elements": [
            {"identifier": "i1", "contents": "0", "class": "File", "tags": ["random", "group:type:single"]},
            {"identifier": "i2", "contents": "1", "class": "File", "tags": ["random", "group:type:paired"]},
            {"identifier": "i3", "contents": "2", "class": "File", "tags": ["random", "group:type:paired"]},
        ],
    },
    "check": check_example_6,
    "output_hid": 8,
}
